"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const cache_1 = require("@teambit/cache");
const cli_1 = require("@teambit/cli");
const component_1 = require("@teambit/component");
const express_1 = require("@teambit/express");
const graphql_1 = require("@teambit/graphql");
const harmony_1 = require("@teambit/harmony");
const logger_1 = require("@teambit/logger");
const pubsub_1 = __importDefault(require("@teambit/pubsub"));
const utils_1 = require("bit-bin/dist/utils");
const fs_extra_1 = __importDefault(require("fs-extra"));
const get_port_1 = __importDefault(require("get-port"));
const path_1 = require("path");
const util_1 = require("util");
const webpack_1 = __importDefault(require("webpack"));
const events_1 = require("./events");
const create_root_1 = require("./create-root");
const exceptions_1 = require("./exceptions");
const start_cmd_1 = require("./start.cmd");
const ui_build_cmd_1 = require("./ui-build.cmd");
const ui_server_1 = require("./ui-server");
const ui_aspect_1 = require("./ui.aspect");
const open_browser_1 = require("./open-browser");
const webpack_config_1 = __importDefault(require("./webpack/webpack.config"));
class UiMain {
    constructor(
    /**
     * Pubsub extension.
     */
    pubsub, config, 
    /**
     * graphql extension.
     */
    graphql, 
    /**
     * slot registry of ui roots.
     */
    uiRootSlot, 
    /**
     * express extension.
     */
    express, 
    /**
     * on start slot
     */
    onStartSlot, 
    /**
     * component extension.
     */
    componentExtension, 
    /**
     * ui logger instance.
     */
    cache, 
    /**
     * ui logger instance.
     */
    logger) {
        this.pubsub = pubsub;
        this.config = config;
        this.graphql = graphql;
        this.uiRootSlot = uiRootSlot;
        this.express = express;
        this.onStartSlot = onStartSlot;
        this.componentExtension = componentExtension;
        this.cache = cache;
        this.logger = logger;
        /**
         * Events
         */
        this.createUiServerStartedEvent = (targetHost, targetPort, uiRoot) => {
            return new events_1.UiServerStartedEvent(Date.now(), targetHost, targetPort, uiRoot);
        };
    }
    /**
     * create a build of the given UI root.
     */
    build(uiRootName) {
        return __awaiter(this, void 0, void 0, function* () {
            const [name, uiRoot] = this.getUi(uiRootName);
            // TODO: @uri refactor all dev server related code to use the bundler extension instead.
            const config = webpack_config_1.default(uiRoot.path, [yield this.generateRoot(yield uiRoot.resolveAspects(ui_aspect_1.UIRuntime.name), name)], uiRoot.name);
            const compiler = webpack_1.default(config);
            const compilerRun = util_1.promisify(compiler.run.bind(compiler));
            return compilerRun();
        });
    }
    /**
     * create a Bit UI runtime.
     */
    createRuntime({ uiRootName, pattern, dev, port, rebuild }) {
        return __awaiter(this, void 0, void 0, function* () {
            const [name, uiRoot] = this.getUi(uiRootName);
            this.componentExtension.setHostPriority(name);
            const uiServer = ui_server_1.UIServer.create({
                express: this.express,
                graphql: this.graphql,
                uiRoot,
                uiRootExtension: name,
                ui: this,
                logger: this.logger,
            });
            const targetPort = yield this.getPort(port);
            if (dev) {
                yield uiServer.dev({ port: targetPort });
            }
            else {
                yield this.buildIfChanged(name, uiRoot, rebuild);
                yield uiServer.start({ port: targetPort });
            }
            // TODO: is this the right place?
            this.pubsub.pub(ui_aspect_1.UIAspect.id, this.createUiServerStartedEvent(this.config.host, targetPort, uiRoot));
            if (uiRoot.postStart)
                yield uiRoot.postStart({ pattern });
            yield this.invokeOnStart();
            // TODO: need to wait until compilation done, then open browser
            // await this.openBrowser(`http://${this.config.host}:${targetPort}`);
            return uiServer;
        });
    }
    getPort(port) {
        return __awaiter(this, void 0, void 0, function* () {
            if (port)
                return port;
            return this.config.port || this.selectPort();
        });
    }
    /**
     * bind to ui server start event.
     */
    registerOnStart(onStartFn) {
        this.onStartSlot.register(onStartFn);
        return this;
    }
    invokeOnStart() {
        return __awaiter(this, void 0, void 0, function* () {
            const promises = this.onStartSlot.values().map((fn) => fn());
            yield Promise.all(promises);
        });
    }
    /**
     * register a UI slot.
     */
    registerUiRoot(uiRoot) {
        return this.uiRootSlot.register(uiRoot);
    }
    /**
     * get a UI runtime instance.
     */
    getUi(uiRootName) {
        if (uiRootName) {
            const root = this.getUiRootOrThrow(uiRootName);
            return [uiRootName, root];
        }
        const uis = this.uiRootSlot.toArray();
        if (uis.length === 1)
            return uis[0];
        const uiRoot = uis.find(([, root]) => root.priority);
        if (!uiRoot)
            throw new exceptions_1.UnknownUI('default');
        return uiRoot;
    }
    getUiRootOrThrow(uiRootName) {
        const uiSlot = this.uiRootSlot.get(uiRootName);
        if (!uiSlot)
            throw new exceptions_1.UnknownUI(uiRootName);
        return uiSlot;
    }
    createLink(aspectDefs, rootExtensionName) {
        return create_root_1.createRoot(aspectDefs, rootExtensionName);
    }
    /**
     * generate the root file of the UI runtime.
     */
    generateRoot(aspectDefs, rootExtensionName, runtimeName = ui_aspect_1.UIRuntime.name, rootAspect = ui_aspect_1.UIAspect.id) {
        return __awaiter(this, void 0, void 0, function* () {
            const contents = yield create_root_1.createRoot(aspectDefs, rootExtensionName, rootAspect, runtimeName);
            const filepath = path_1.resolve(path_1.join(__dirname, `${runtimeName}.root${utils_1.sha1(contents)}.js`));
            if (fs_extra_1.default.existsSync(filepath))
                return filepath;
            fs_extra_1.default.outputFileSync(filepath, contents);
            return filepath;
        });
    }
    selectPort() {
        return __awaiter(this, void 0, void 0, function* () {
            const [from, to] = this.config.portRange;
            return get_port_1.default({ port: get_port_1.default.makeRange(from, to) });
        });
    }
    buildUiHash(uiRoot, runtime = 'ui') {
        return __awaiter(this, void 0, void 0, function* () {
            const aspects = yield uiRoot.resolveAspects(runtime);
            aspects.sort((a, b) => (a.aspectPath > b.aspectPath ? 1 : -1));
            const hash = aspects.map((aspect) => {
                return [aspect.aspectPath, aspect.runtimePath].join('');
            });
            return utils_1.sha1(hash.join(''));
        });
    }
    buildIfChanged(name, uiRoot, force) {
        return __awaiter(this, void 0, void 0, function* () {
            const hash = yield this.buildUiHash(uiRoot);
            const hashed = yield this.cache.get(uiRoot.path);
            if (hash === hashed && !force)
                return;
            if (hash !== hashed)
                this.logger.console(`${uiRoot.configFile} has been changed. Rebuilding UI assets for ${uiRoot.name}`);
            this.logger.console(`Building UI assets for ${uiRoot.name}`);
            yield this.build(name);
            yield this.cache.set(uiRoot.path, hash);
        });
    }
    openBrowser(url) {
        return __awaiter(this, void 0, void 0, function* () {
            const openBrowser = new open_browser_1.OpenBrowser(this.logger);
            openBrowser.open(url);
        });
    }
    static provider([pubsub, cli, graphql, express, componentExtension, cache, loggerMain], config, [uiRootSlot, onStartSlot]) {
        return __awaiter(this, void 0, void 0, function* () {
            // aspectExtension.registerRuntime(new RuntimeDefinition('ui', []))
            const logger = loggerMain.createLogger(ui_aspect_1.UIAspect.id);
            const ui = new UiMain(pubsub, config, graphql, uiRootSlot, express, onStartSlot, componentExtension, cache, logger);
            cli.register(new start_cmd_1.StartCmd(ui, logger, pubsub));
            cli.register(new ui_build_cmd_1.UIBuildCmd(ui));
            return ui;
        });
    }
}
exports.UiMain = UiMain;
UiMain.defaultConfig = {
    portRange: [3000, 3100],
    host: 'localhost',
};
UiMain.runtime = cli_1.MainRuntime;
UiMain.dependencies = [
    pubsub_1.default,
    cli_1.CLIAspect,
    graphql_1.GraphqlAspect,
    express_1.ExpressAspect,
    component_1.ComponentAspect,
    cache_1.CacheAspect,
    logger_1.LoggerAspect,
];
UiMain.slots = [harmony_1.Slot.withType(), harmony_1.Slot.withType()];
ui_aspect_1.UIAspect.addRuntime(UiMain);
//# sourceMappingURL=ui.main.runtime.js.map